/*!
 *  \file      mmc_avr.c
 *  \brief     This is customized version of mmc_avr.c 
 *             The original version is from sample \e avr release R0.11a
 *  \date      11 march 2020 (previous released 2016)
 *          
 *  \details   It is modified by Wim Dolman (<a href="mailto:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>) 
 *             for the HvA-Xmegaboard version 2
 *             The most important global changes are:
 *             - It is suitable for the ATxmega256a3u.
 *             - It uses the Xmega style (DIRSET, OUTSET, ...).
 *             - It uses SPIC with a remap of (SCK and MOSI) in stead of bit banging.
 *             .
 *  
 *             The most important detailed changes are:
 *             1.  A static function sdspi_init() is added that initialize the SPI
 *             2.  A static function sdspi_write_multi_byte() is added to write multiple bytes to the SPI
 *             3.  A static function sdspi_read_multi_byte() is added to read multiple bytes from the SPI
 *             4.  The static function power_on() is changed. It initislizes the SPI of port C in the Xmega-style.
 *                 The power dectection is removed.
 *             5.  The static function power_off() is changed. It disables the SPI of port C in the Xmega-style.
 *                 The power detcetion is removed.
 *             6.  The static function xchg_spi() is now an alias for sdspi_transfer_byte()
 *             7.  The static function xmit_spi_multi() is now an alias for sdspi_write_multi_byte()
 *             8.  The static function rcvr_spi_multi() is now an alias for sdspi_read_multi_byte()
 *
 */
 
/*-----------------------------------------------------------------------*/
/* MMCv3/SDv1/SDv2 (in SPI mode) control module                          */
/*-----------------------------------------------------------------------*/
/*
/  Copyright (C) 2014, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/  Changed 30 march 2016
/  Wim Dolman (<a href="email:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>)
/  Made it suitable for the HvA-Xmegaboard. The most important chenges are:
/    -  Xmega256au3
/    -  Xmega-style (DIRSET, OUTSET, ...)
/    -  Using SPIC with remap of (SCK and MOSI) in stead of SPDR
/    -  Socket power control removed. HvA-Xmegaboard has no power control.
/
/  This implentation uses a SPI and a timeout. For the timeout a
/  100Hz decrement timer is necessary. For example timer D0:
/    void init_timer(void)
/    {
/      TCD0.CTRLB     = TC_WGMODE_NORMAL_gc;
/      TCD0.CTRLA     = TC_CLKSEL_DIV8_gc;
/      TCD0.INTCTRLA  = TC_OVFINTLVL_LO_gc;
/      TCD0.PER       = 39999;  // f=fcpu/(N*(PER+1))=32M/8*40000=100 Hz
/    }
/
/    ISR(TCD0_OVF_vect)
/    {
/       disk_timerproc();  // Drive timer procedure of low level disk I/O
/    }
/-------------------------------------------------------------------------*/
//! \cond CHANGED1

#include <avr/io.h>
#include "diskio.h"
//! \endcond

//! \cond FROMRELEASE12
/*--------------------------------------------------------------------------
   Module Private Functions
---------------------------------------------------------------------------*/
//@cond
// MMC/SD command (SPI mode)
#define CMD0   (0)        /* GO_IDLE_STATE */
#define CMD1   (1)        /* SEND_OP_COND */
#define ACMD41 (0x80+41)  /* SEND_OP_COND (SDC) */
#define CMD8   (8)        /* SEND_IF_COND */
#define CMD9   (9)        /* SEND_CSD */
#define CMD10  (10)       /* SEND_CID */
#define CMD12  (12)       /* STOP_TRANSMISSION */
#define CMD13  (13)       /* SEND_STATUS */
#define ACMD13 (0x80+13)  /* SD_STATUS (SDC) */
#define CMD16  (16)       /* SET_BLOCKLEN */
#define CMD17  (17)       /* READ_SINGLE_BLOCK */
#define CMD18  (18)       /* READ_MULTIPLE_BLOCK */
#define CMD23  (23)       /* SET_BLOCK_COUNT */
#define ACMD23 (0x80+23)  /* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24  (24)       /* WRITE_BLOCK */
#define CMD25  (25)       /* WRITE_MULTIPLE_BLOCK */
#define CMD32  (32)       /* ERASE_ER_BLK_START */
#define CMD33  (33)       /* ERASE_ER_BLK_END */
#define CMD38  (38)       /* ERASE */
#define	CMD48	 (48)       /* READ_EXTR_SINGLE */
#define	CMD49  (49)       /* WRITE_EXTR_SINGLE */
#define CMD55  (55)       /* APP_CMD */
#define CMD58  (58)       /* READ_OCR */

static DSTATUS Stat = STA_NOINIT;     // Disk status
static BYTE CardType;                 // b0:MMC, b1:SDv1, b2:SDv2,
                                      //         b3:Block addressing
static volatile BYTE Timer1, Timer2;  // 100Hz decrement timer
//! \endcond

//! \cond CHANGED2
/*-----------------------------------------------------------------------*/
/* SPI Chip Select for SD-card HvA-Xmegaboard version 2                  */
/*-----------------------------------------------------------------------*/
#define sdspi_CSlow()   PORTC.OUTCLR = PIN4_bm           // added:  define for Chip Select low  
#define sdspi_CShigh()  PORTC.OUTSET = PIN4_bm           // added:  define for Chip Select high
#define FCLK_SLOW()     SPIC.CTRL = (SPI_CLK2X_bm|SPI_ENABLE_bm|SPI_MASTER_bm|SPI_MODE_0_gc|SPI_PRESCALER_DIV64_gc)  // added:  set slow clock
#define FCLK_FAST()     SPIC.CTRL = (SPI_CLK2X_bm|SPI_ENABLE_bm|SPI_MASTER_bm|SPI_MODE_0_gc|SPI_PRESCALER_DIV4_gc)   // added:  set fast clock
#define CS_LOW()        PORTC.OUTCLR = PIN4_bm           // added:  define for Chip Select low 
#define CS_HIGH()       PORTC.OUTSET = PIN4_bm           // added:  define for Chip Select high
#define MMC_CD          ( bit_is_clear(PORTF, 4) )       // Card detected. yes:non zero, no:zero
#define MMC_WP          0                                // Write protected not implemented with HvA-Xmegaboard

/*-----------------------------------------------------------------------*/
/* Transmit/Receive data from/to MMC via SPI                             */
/*-----------------------------------------------------------------------*/
static unsigned char sdspi_transfer_byte(unsigned char databyte)
{
  SPIC.DATA = databyte;
  while ( !(SPIC.STATUS & (SPI_IF_bm))) ;
  return SPIC.DATA;
}

/*-----------------------------------------------------------------------*/
/* Transmit data block to MMC via SPI                                    */
/*-----------------------------------------------------------------------*/
static void sdspi_write_multi_byte(const unsigned char *p, unsigned int cnt)
{
  do {
    SPIC.DATA = *p++;
    while ( !(SPIC.STATUS & (SPI_IF_bm))) ;
    SPIC.DATA = *p++;
    while ( !(SPIC.STATUS & (SPI_IF_bm))) ;
  } while (cnt -= 2);
}

/*-----------------------------------------------------------------------*/
/* Receive data block from MMC via SPI                                   */
/*-----------------------------------------------------------------------*/
static void sdspi_read_multi_byte(unsigned char *p, unsigned int cnt)
{
  do {
    SPIC.DATA = 0xFF;
    while ( !(SPIC.STATUS & (SPI_IF_bm))) ;
    *p++ = SPIC.DATA;
    SPIC.DATA = 0xFF;
    while ( !(SPIC.STATUS & (SPI_IF_bm))) ;
    *p++ = SPIC.DATA;
  } while (cnt -= 2);
}

/*-----------------------------------------------------------------------*/
/* Replacements voor spi-functions                                       */
/*-----------------------------------------------------------------------*/
#define xchg_spi        sdspi_transfer_byte                  // alias for sdspi_transfer_byte()
#define xmit_spi_multi  sdspi_write_multi_byte               // alias for sdspi_write_multi_byte()
#define rcvr_spi_multi  sdspi_read_multi_byte                // alias for sdspi_read_multi_byte()

/*-----------------------------------------------------------------------*/
/* Power Control  (Platform dependent)                                   */
/*-----------------------------------------------------------------------*/
/* When the target system does not support socket power control, there   */
/* is nothing to do in these functions and chk_power always returns 1.   */

static void power_on (void)
{
  // {  /* Remove this block if no socket power control */
    //PORTE &= ~_BV(7);  /* Socket power on (PE7=low) */
    //DDRE |= _BV(7);
    //for (Timer1 = 2; Timer1; );  /* Wait for 20ms */
  // }

  PORTC.REMAP    = PORT_SPI_bm;         // swap SCK and MOSI
  PORTC.DIRCLR   = PIN6_bm;             // MISO        input
  PORTC.DIRSET   = PIN5_bm;             // SCK         output
  PORTC.DIRSET   = PIN7_bm;             // MOSI        output
  PORTC.DIRSET   = PIN4_bm;             // chip select output
  PORTC.PIN6CTRL = PORT_OPC_PULLUP_gc;  // MISO        pullup

  PORTF.DIRCLR   = PIN4_bm;             // card detect input
  PORTF.PIN4CTRL = PORT_OPC_PULLUP_gc;  // card detect pullup

// was getest  SPIC.CTRL = (SPI_CLK2X_bm|SPI_ENABLE_bm|SPI_MASTER_bm|SPI_MODE_0_gc|SPI_PRESCALER_DIV64_gc);
// 8 MHz misschien kan 16 ook
  SPIC.CTRL = SPI_ENABLE_bm|SPI_MASTER_bm|SPI_MODE_0_gc|SPI_PRESCALER_DIV4_gc;
}

static void power_off (void)
{
  SPIC.CTRL = 0;       // Disable SPI function

  PORTC.REMAP    = PORT_SPI_bm;         // swap SCK and MOSI
  PORTC.DIRCLR   = PIN6_bm;             // MISO        input
  PORTC.DIRSET   = PIN5_bm;             // SCK         output
  PORTC.DIRSET   = PIN7_bm;             // MOSI        output
  PORTC.DIRSET   = PIN4_bm;             // chip select output
  PORTC.PIN6CTRL = PORT_OPC_PULLUP_gc;  // MISO        pullup

  PORTF.DIRCLR   = PIN4_bm;             // card detect input
  PORTF.PIN4CTRL = PORT_OPC_PULLUP_gc;  // card detect pullup

  {  /* Remove this block if no socket power control */
    // PORTE |= _BV(7);    /* Socket power off (PE7=high) */
    // for (Timer1 = 20; Timer1; );  /* Wait for 20ms */
  }
}
//! \endcond

//! \cond NOTCHANGED
/*
 * FROM HERE NOTHING CHANGED!
 * From it is completely equal to avr-mmc.c of example avr/mmc-avr-spi.c 
 * from release 12.
 * Except the headers from the disk-functies are from mmc-avr.c release 11a.
 * 
 */


/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/

static
int wait_ready (  /* 1:Ready, 0:Timeout */
  UINT wt      /* Timeout [ms] */
)
{
  BYTE d;

  Timer2 = wt / 10;
  do
    d = xchg_spi(0xFF);
  while (d != 0xFF && Timer2);

  return (d == 0xFF) ? 1 : 0;
}



/*-----------------------------------------------------------------------*/
/* Deselect the card and release SPI bus                                 */
/*-----------------------------------------------------------------------*/

static
void deselect (void)
{
  CS_HIGH();    /* Set CS# high */
  xchg_spi(0xFF);  /* Dummy clock (force DO hi-z for multiple slave SPI) */
}



/*-----------------------------------------------------------------------*/
/* Select the card and wait for ready                                    */
/*-----------------------------------------------------------------------*/

static
int select (void)  /* 1:Successful, 0:Timeout */
{
  CS_LOW();    /* Set CS# low */
  xchg_spi(0xFF);  /* Dummy clock (force DO enabled) */
  if (wait_ready(500)) return 1;  /* Wait for card ready */

  deselect();
  return 0;  /* Timeout */
}



/*-----------------------------------------------------------------------*/
/* Receive a data packet from MMC                                        */
/*-----------------------------------------------------------------------*/

static
int rcvr_datablock (
  BYTE *buff,      /* Data buffer to store received data */
  UINT btr      /* Byte count (must be multiple of 4) */
)
{
  BYTE token;


  Timer1 = 20;
  do {              /* Wait for data packet in timeout of 200ms */
    token = xchg_spi(0xFF);
  } while ((token == 0xFF) && Timer1);
  if (token != 0xFE) return 0;  /* If not valid data token, retutn with error */

  rcvr_spi_multi(buff, btr);    /* Receive the data block into buffer */
  xchg_spi(0xFF);          /* Discard CRC */
  xchg_spi(0xFF);

  return 1;            /* Return with success */
}



/*-----------------------------------------------------------------------*/
/* Send a data packet to MMC                                             */
/*-----------------------------------------------------------------------*/

#if  _USE_WRITE
static
int xmit_datablock (
  const BYTE *buff,  /* 512 byte data block to be transmitted */
  BYTE token      /* Data/Stop token */
)
{
  BYTE resp;


  if (!wait_ready(500)) return 0;

  xchg_spi(token);          /* Xmit data token */
  if (token != 0xFD) {  /* Is data token */
    xmit_spi_multi(buff, 512);    /* Xmit the data block to the MMC */
    xchg_spi(0xFF);          /* CRC (Dummy) */
    xchg_spi(0xFF);
    resp = xchg_spi(0xFF);      /* Reveive data response */
    if ((resp & 0x1F) != 0x05)    /* If not accepted, return with error */
      return 0;
  }

  return 1;
}
#endif



/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/

static
BYTE send_cmd (    /* Returns R1 resp (bit7==1:Send failed) */
  BYTE cmd,    /* Command index */
  DWORD arg    /* Argument */
)
{
  BYTE n, res;


  if (cmd & 0x80) {  /* ACMD<n> is the command sequense of CMD55-CMD<n> */
    cmd &= 0x7F;
    res = send_cmd(CMD55, 0);
    if (res > 1) return res;
  }

  /* Select the card and wait for ready except to stop multiple block read */
  if (cmd != CMD12) {
    deselect();
    if (!select()) return 0xFF;
  }

  /* Send command packet */
  xchg_spi(0x40 | cmd);        /* Start + Command index */
  xchg_spi((BYTE)(arg >> 24));    /* Argument[31..24] */
  xchg_spi((BYTE)(arg >> 16));    /* Argument[23..16] */
  xchg_spi((BYTE)(arg >> 8));      /* Argument[15..8] */
  xchg_spi((BYTE)arg);        /* Argument[7..0] */
  n = 0x01;              /* Dummy CRC + Stop */
  if (cmd == CMD0) n = 0x95;      /* Valid CRC for CMD0(0) + Stop */
  if (cmd == CMD8) n = 0x87;      /* Valid CRC for CMD8(0x1AA) Stop */
  xchg_spi(n);

  /* Receive command response */
  if (cmd == CMD12) xchg_spi(0xFF);    /* Skip a stuff byte when stop reading */
  n = 10;                /* Wait for a valid response in timeout of 10 attempts */
  do
    res = xchg_spi(0xFF);
  while ((res & 0x80) && --n);

  return res;      /* Return with the response value */
}



/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
  BYTE pdrv    /* Physical drive nmuber (0) */
)
{
  BYTE n, cmd, ty, ocr[4];


  if (pdrv) return STA_NOINIT;    /* Supports only single drive */
	for (Timer1 = 10; Timer1; ) ;		/* Wait for 100ms */
	if (Stat & STA_NODISK) return Stat;	/* No card in the socket? */

	power_on();							/* Turn on the socket power */
	FCLK_SLOW();
	for (n = 10; n; n--) xchg_spi(0xFF);	/* 80 dummy clocks */

	ty = 0;
	if (send_cmd(CMD0, 0) == 1) {			/* Put the card SPI mode */
		Timer1 = 100;						/* Initialization timeout of 1000 msec */
		if (send_cmd(CMD8, 0x1AA) == 1) {	/* Is the card SDv2? */
			for (n = 0; n < 4; n++) ocr[n] = xchg_spi(0xFF);	/* Get trailing return value of R7 resp */
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) {				/* The card can work at vdd range of 2.7-3.6V */
				while (Timer1 && send_cmd(ACMD41, 1UL << 30));	/* Wait for leaving idle state (ACMD41 with HCS bit) */
				if (Timer1 && send_cmd(CMD58, 0) == 0) {		/* Check CCS bit in the OCR */
					for (n = 0; n < 4; n++) ocr[n] = xchg_spi(0xFF);
					ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;	/* Check if the card is SDv2 */
				}
			}
		} else {							/* SDv1 or MMCv3 */
			if (send_cmd(ACMD41, 0) <= 1) 	{
				ty = CT_SD1; cmd = ACMD41;	/* SDv1 */
			} else {
				ty = CT_MMC; cmd = CMD1;	/* MMCv3 */
			}
			while (Timer1 && send_cmd(cmd, 0));			/* Wait for leaving idle state */
			if (!Timer1 || send_cmd(CMD16, 512) != 0)	/* Set R/W block length to 512 */
				ty = 0;
		}
	}
	CardType = ty;
	deselect();

	if (ty) {			/* Initialization succeded */
		Stat &= ~STA_NOINIT;		/* Clear STA_NOINIT */
		FCLK_FAST();
	} else {			/* Initialization failed */
		power_off();
	}

	return Stat;}



/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
  BYTE pdrv    /* Physical drive nmuber (0) */
)
{
  if (pdrv) return STA_NOINIT;  /* Supports only single drive */
  return Stat;
}



/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
  BYTE pdrv,      /* Physical drive nmuber (0) */
  BYTE *buff,      /* Pointer to the data buffer to store read data */
  DWORD sector,    /* Start sector number (LBA) */
  UINT count      /* Sector count (1..128) */
)
{
  BYTE cmd;
	DWORD sect = (DWORD)sector;


  if (pdrv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;

	if (!(CardType & CT_BLOCK)) sect *= 512;	/* Convert to byte address if needed */

	cmd = count > 1 ? CMD18 : CMD17;			/*  READ_MULTIPLE_BLOCK : READ_SINGLE_BLOCK */
	if (send_cmd(cmd, sect) == 0) {
		do {
			if (!rcvr_datablock(buff, 512)) break;
			buff += 512;
		} while (--count);
		if (cmd == CMD18) send_cmd(CMD12, 0);	/* STOP_TRANSMISSION */
	}
	deselect();

	return count ? RES_ERROR : RES_OK;
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _USE_WRITE
DRESULT disk_write (
  BYTE pdrv,      /* Physical drive nmuber (0) */
  const BYTE *buff,  /* Pointer to the data to be written */
  DWORD sector,    /* Start sector number (LBA) */
  UINT count      /* Sector count (1..128) */
)
{
	DWORD sect = (DWORD)sector;

  
  if (pdrv || !count) return RES_PARERR;
  if (Stat & STA_NOINIT) return RES_NOTRDY;
  if (Stat & STA_PROTECT) return RES_WRPRT;

	if (!(CardType & CT_BLOCK)) sect *= 512;	/* Convert to byte address if needed */

	if (count == 1) {	/* Single block write */
		if ((send_cmd(CMD24, sect) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE)) {
			count = 0;
		}
	}
	else {				/* Multiple block write */
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sect) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD)) count = 1;	/* STOP_TRAN token */
		}
	}
	deselect();

	return count ? RES_ERROR : RES_OK;
}
#endif


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL
DRESULT disk_ioctl (
  BYTE pdrv,    /* Physical drive nmuber (0) */
  BYTE cmd,    /* Control code */
  void *buff    /* Buffer to send/receive control data */
)
{
  DRESULT res;
  BYTE n, csd[16], *ptr = buff;
  DWORD csize;
#if FF_USE_TRIM
	LBA_t *range;
	DWORD st, ed;
#endif
#if _USE_ISDIO
	SDIO_CTRL *sdi;
	BYTE rc, *bp;
	UINT dc;
#endif


  if (pdrv) return RES_PARERR;

	if (Stat & STA_NOINIT) return RES_NOTRDY;

	res = RES_ERROR;
	switch (cmd) {
	case CTRL_SYNC :		/* Make sure that no pending write process. Do not remove this or written sector might not left updated. */
		if (select()) res = RES_OK;
		deselect();
		break;

	case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (DWORD) */
		if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
			if ((csd[0] >> 6) == 1) {	/* SDC ver 2.00 */
				csize = csd[9] + ((WORD)csd[8] << 8) + ((DWORD)(csd[7] & 63) << 16) + 1;
				*(LBA_t*)buff = csize << 10;
			} else {					/* SDC ver 1.XX or MMC*/
				n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
				csize = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
				*(LBA_t*)buff = csize << (n - 9);
			}
			res = RES_OK;
		}
		deselect();
		break;

	case GET_BLOCK_SIZE :	/* Get erase block size in unit of sector (DWORD) */
		if (CardType & CT_SD2) {	/* SDv2? */
			if (send_cmd(ACMD13, 0) == 0) {	/* Read SD status */
				xchg_spi(0xFF);
				if (rcvr_datablock(csd, 16)) {				/* Read partial block */
					for (n = 64 - 16; n; n--) xchg_spi(0xFF);	/* Purge trailing data */
					*(DWORD*)buff = 16UL << (csd[10] >> 4);
					res = RES_OK;
				}
			}
		} else {					/* SDv1 or MMCv3 */
			if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {	/* Read CSD */
				if (CardType & CT_SD1) {	/* SDv1 */
					*(DWORD*)buff = (((csd[10] & 63) << 1) + ((WORD)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
				} else {					/* MMCv3 */
					*(DWORD*)buff = ((WORD)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
				}
				res = RES_OK;
			}
		}
		deselect();
		break;
#if FF_USE_TRIM
	case CTRL_TRIM:		/* Erase a block of sectors (used when _USE_TRIM in ffconf.h is 1) */
		if (!(CardType & CT_SDC)) break;				/* Check if the card is SDC */
		if (mmc_disk_ioctl(MMC_GET_CSD, csd)) break;	/* Get CSD */
		if (!(csd[0] >> 6) && !(csd[10] & 0x40)) break;	/* Check if sector erase can be applied to the card */
		range = buff; st = (DWORD)range[0]; ed = (DWORD)range[1];	/* Load sector block */
		if (!(CardType & CT_BLOCK)) {
			st *= 512; ed *= 512;
		}
		if (send_cmd(CMD32, st) == 0 && send_cmd(CMD33, ed) == 0 && send_cmd(CMD38, 0) == 0 && wait_ready(60000)) {	/* Erase sector block */
			res = RES_OK;	/* FatFs does not check result of this command */
		}
		break;
#endif
	/* Following commands are never used by FatFs module */

	case MMC_GET_TYPE :		/* Get card type flags (1 byte) */
		*ptr = CardType;
		res = RES_OK;
		break;

	case MMC_GET_CSD :		/* Receive CSD as a data block (16 bytes) */
		if (send_cmd(CMD9, 0) == 0 && rcvr_datablock(ptr, 16)) {	/* READ_CSD */
			res = RES_OK;
		}
		deselect();
		break;

	case MMC_GET_CID :		/* Receive CID as a data block (16 bytes) */
		if (send_cmd(CMD10, 0) == 0 && rcvr_datablock(ptr, 16)) {	/* READ_CID */
			res = RES_OK;
		}
		deselect();
		break;

	case MMC_GET_OCR :		/* Receive OCR as an R3 resp (4 bytes) */
		if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
			for (n = 4; n; n--) *ptr++ = xchg_spi(0xFF);
			res = RES_OK;
		}
		deselect();
		break;

	case MMC_GET_SDSTAT :	/* Receive SD statsu as a data block (64 bytes) */
		if (send_cmd(ACMD13, 0) == 0) {	/* SD_STATUS */
			xchg_spi(0xFF);
			if (rcvr_datablock(ptr, 64)) res = RES_OK;
		}
		deselect();
		break;

	case CTRL_POWER_OFF :	/* Power off */
		power_off();
		Stat |= STA_NOINIT;
		res = RES_OK;
		break;
#if _USE_ISDIO
	case ISDIO_READ:
		sdi = buff;
		if (send_cmd(CMD48, 0x80000000 | (DWORD)sdi->func << 28 | (DWORD)sdi->addr << 9 | ((sdi->ndata - 1) & 0x1FF)) == 0) {
			for (Timer1 = 100; (rc = xchg_spi(0xFF)) == 0xFF && Timer1; ) ;
			if (rc == 0xFE) {
				for (bp = sdi->data, dc = sdi->ndata; dc; dc--) *bp++ = xchg_spi(0xFF);
				for (dc = 514 - sdi->ndata; dc; dc--) xchg_spi(0xFF);
				res = RES_OK;
			}
		}
		deselect();
		break;

	case ISDIO_WRITE:
		sdi = buff;
		if (send_cmd(CMD49, 0x80000000 | (DWORD)sdi->func << 28 | (DWORD)sdi->addr << 9 | ((sdi->ndata - 1) & 0x1FF)) == 0) {
			xchg_spi(0xFF); xchg_spi(0xFE);
			for (bp = sdi->data, dc = sdi->ndata; dc; dc--) xchg_spi(*bp++);
			for (dc = 514 - sdi->ndata; dc; dc--) xchg_spi(0xFF);
			if ((xchg_spi(0xFF) & 0x1F) == 0x05) res = RES_OK;
		}
		deselect();
		break;

	case ISDIO_MRITE:
		sdi = buff;
		if (send_cmd(CMD49, 0x84000000 | (DWORD)sdi->func << 28 | (DWORD)sdi->addr << 9 | sdi->ndata >> 8) == 0) {
			xchg_spi(0xFF); xchg_spi(0xFE);
			xchg_spi(sdi->ndata);
			for (dc = 513; dc; dc--) xchg_spi(0xFF);
			if ((xchg_spi(0xFF) & 0x1F) == 0x05) res = RES_OK;
		}
		deselect();
		break;
#endif
	default:
		res = RES_PARERR;
	}

	return res;
}
#endif


/*-----------------------------------------------------------------------*/
/* Device Timer Interrupt Procedure                                      */
/*-----------------------------------------------------------------------*/
/* This function must be called in period of 10ms                        */

void disk_timerproc (void)
{
	BYTE b;
	UINT n;


	b = Timer1;				/* 100Hz decrement timer */
	if (b) Timer1 = --b;
	n = Timer2;
	if (n) Timer2 = --n;

	b = Stat;
	if (MMC_WP) {				/* Write protected */
		b |= STA_PROTECT;
	} else {					/* Write enabled */
		b &= ~STA_PROTECT;
	}
	if (MMC_CD) {				/* Card inserted */
		b &= ~STA_NODISK;
	} else {					/* Socket empty */
		b |= (STA_NODISK | STA_NOINIT);
	}
	Stat = b;				/* Update MMC status */
}
//! \endcond
