/*!
 *  \file    sdcard_test_mmc_avr.c
 *  \author  Wim Dolman (<a href="mailto:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>)
 *  \date    11-03-2020
 *  \version 1.5
 *
 *  \brief   Test program for sd-card with Chan FatFS and customized mmc_avr.c
 *
 *  \details This implementation needs a timer for the necessary delays.
 *           <br><br>
 *           <b>Usage with Atmel Studio.</b><br>
 *           This example uses the serialF0, see https://dolman-wim.nl/xmega/libraries/online/serialF0/html/index.html
 *           and the clock functions, see https://dolman-wim.nl/xmega/libraries/online/clock/html/index.html
 *           In the project directory uses three subfolders: clock, ff, and serialF0.
 *           <ol><li>Make an Atmel Studio project
 *           </li><li>Add three subfolders (use New Folder...) to the project  : clock, ff, and serialF0
 *           </li><li>Place the necessary files in the Atmel Studio project folder/subfolders
 *               <table>
 *               <tr><td>clock/clock.c          </td><td>Clock functions</td></tr>
 *               <tr><td>clock/clock.h          </td><td>Clock functions</td></tr>
 *               <tr><td>serialF0/serialF0.c    </td><td>Serial interface</td></tr>
 *               <tr><td>serialF0/serialF0.h    </td><td>Serial interface</td></tr>
 *               <tr><td>ff/ff.c                </td><td>Latest version from Chan's Library</td></tr>
 *               <tr><td>ff/ff.h                </td><td>Latest version from Chan's Library</td></tr>
 *               <tr><td>ff/ffconf.h            </td><td>Latest version from Chan's Library</td></tr>
 *               <tr><td>ff/diskio.h            </td><td>Version 2014 from Chan's \e avr releases 11a</td></tr>
 *               <tr><td>ff/ mmc_avr.c          </td><td>Customized version Chan's Sample \e avr releases 11a/12</td></tr>
 *               <tr><td>sdcard_test_mmc_avr.c  </td><td>This example</td></tr>
 *               <tr><td>demo_ff.c              </td><td>function with demo SD-card</td></tr>
 *               </table>
 *           </li><li> Add all c-files to the project
 *           </li><li> Change these four defines in ffconf.h
 *           \verbatim
               #define FF_USE_STRFUNC  2
               #define FF_USE_CHMO     1
               #define FF_FS_NORTC     1
               #define FF_USE_LFN      0 \endverbatim
 *           </li></ol>
 *
 *           You can place the library files also directly in the project folder or in an other
 *           folder. In that case you need to change the path in the <code>\#include</code> lines.
 *
 *           This test program is based on the example D.2 and D.3 from
 *           <a href="http://www.dolman-wim.nl/xmega/index.php">'De taal C en de Xmega'</a> 2nd edition.
 *
 */
#ifndef F_CPU
#define F_CPU 32000000UL
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include "ff/ff.h"
#include "ff/diskio.h"
#include "clock/clock.h"
#include "serialF0/serialF0.h"

void init_timer(void);
void demo_ff(void);

/*! \brief   main routine for testting sdmm.c
 *
 *           It initialize the clock, the streams, the toimer,
 *           set the global interrupts and run the demo program once.
 *
 *  \return  int
 */
int main (void)
{
  init_clock();
  init_stream(F_CPU);

  init_timer();
  PMIC.CTRL |= PMIC_LOLVLEN_bm;
  sei();

  demo_ff();

  while(1) {}  // do nothing
}

/*! \brief   Initialize timer/counter C1 as  100 Hz timer
 *
 *  \return  void
 */
void init_timer(void)
{
  TCC1.CTRLB     = TC_WGMODE_NORMAL_gc;
  TCC1.CTRLA     = TC_CLKSEL_DIV8_gc;
  TCC1.INTCTRLA  = TC_OVFINTLVL_LO_gc;
  TCC1.PER       = 39999;
}

/*! \brief   Timer Interrupt that triggers disk_timerproc()
 *
 *  \return  void
 */
ISR(TCC1_OVF_vect)
{
  disk_timerproc();  // Update timerinfo, see mmc_avr.c
}
